/*
 * Copyright (C) 2009 Castrense Savojardo
 * 	Bologna Biocomputing Group
 * 	University of Bologna, Italy
 *  savojard@biocomp.unibo.it
 * 
 * ForwardBackward.cpp - This file is part of BioCRF
 * 
 * BioCRF is a free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * BioCRF is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BioCRF; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 */
#include <cmath>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include "ForwardBackward.h"
#include "Data.h"
#include "FeatureSet.h"
#include "Model.h"

namespace algo {
ForwardBackward::ForwardBackward(int seq_len, int num_states)
{
	this->seq_len = seq_len;

	this->num_states = num_states;
	this->index = 0;
	this->clamped = false;
	this->scaling = true;
	this->sum_on_labels = false;
	this->bos = true;
	this->eos = true;
	
	this->scale = new double[this->seq_len];
	
	this->beg_idx = 0;
	this->end_idx = 1;
	
	this->bw_m = new mathlib::DoubleMatrix(num_states, seq_len);
	
	this->fw = new mathlib::DoubleVector(num_states);
	this->next_fw = new mathlib::DoubleVector(num_states);
	
	this->Mi = new mathlib::DoubleMatrix(num_states, num_states);
	this->Vi = new mathlib::DoubleVector(num_states);
	
}

ForwardBackward::~ForwardBackward()
{	
	
	
	if (bw_m != NULL)
		delete bw_m;
	if (scale != NULL)
		delete [] scale;
	if (Mi != NULL)
		delete Mi;
	if (fw != NULL)
		delete fw;
	if (next_fw != NULL)
		delete next_fw;
	if (Vi != NULL)
		delete Vi;	
}

void ForwardBackward::reset(int seq_len)
{
	delete bw_m;
	delete [] scale;
	this->seq_len = seq_len;
	
	this->bw_m = new mathlib::DoubleMatrix(this->num_states, seq_len);
	this->scale = new double[this->seq_len];
	
}

void ForwardBackward::run(data::Sequence & seq, models::Model * model, double * weights)
{
	int i;
	std::ofstream fs;
	char buffer[512];

	run_backward(seq, model, weights);
	
	start_forward_scan();
	
	if (model->options->out_posterior) {
		fs.open(model->options->posterior_out_file.c_str(), std::ios::app);
		if (! fs.is_open()) {
			std::cout << "Error opening posterior output file." << std::endl;
			exit(EXIT_FAILURE);
		}
		for (int j=0; j<model->transducer->num_states; j++){
			std::map<int,std::string>::iterator it = model->transducer->dictstint->find(j);
			fs << it->second << " ";
		}
		fs << std::endl;
	}
							 

	for (i=0; i<seq_len; i++) {
		//bw_m->column_multiply(i, next_fw, 0);

		//bw_m->column_multiply(i, scale[i]/Zx);
		
		for (int j=0; j<num_states; j++)
			bw_m->set(j, i, p(j));
		
		if (model->options->out_posterior) {
	        	for (int j=0; j<model->transducer->num_states; j++)
			{
				sprintf(buffer, "%1.4f", bw_m->get(j,i));
				std::string v = buffer;
				fs << v << " ";
			}
			
			fs << std::endl;
		}
	
		double sums[model->transducer->num_labels];
		if (this->sum_on_labels) {
			int j;
			std::map<int,int>::iterator slmit;
			for (j=0; j<model->transducer->num_labels; j++)
				sums[j] = 0.0;
			for (j=0; j<num_states; j++)
				sums[model->transducer->stlblmap->find(j)->second] += bw_m->get(j,i);
			for (j=0; j<num_states; j++)
				bw_m->set(j, i, sums[model->transducer->stlblmap->find(j)->second]);
		}
		take_forward_step(seq, model, weights);
	}
        if (model->options->out_posterior) {
		fs << std::endl;
		fs.close();
	}
}

void ForwardBackward::run_backward(data::Sequence & seq, models::Model * model, double * weights)
{
	int i;
	
	reset(seq.size());
	
	bw_m->set_column(seq_len-1, 0.0);
	bw_m->set(end_idx, seq_len-1, 1.0);
	scale[seq_len-1] = 1.0;
	
	seq[seq_len-1].compute_transition_feature_scores(model->featureset, Mi, weights);
	
	if (! eos) 
		Mi->set_column(end_idx, 1.0);
	if (! bos)
		Mi->set_row(beg_idx, 1.0);
	
	Mi->multiply_entrywise(model->transducer->transitions, 0);
	
	for (i=seq_len-2; i>=0; i--) {
		seq[i+1].compute_state_feature_scores(model->featureset, seq, i+1, Vi, weights);
		
		mathlib::DoubleVector temp(num_states);
		bw_m->get_column(&temp, i+1);
		temp.multiply(Vi);
			
		Mi->multiply(bw_m, i, &temp, 0);
		
		if (! clamped) {
		if (i == 0) {
			double tmp = bw_m->get(beg_idx, 0); 
			bw_m->set_column(0, 0.0);
			bw_m->set(beg_idx, 0, tmp);
		} else {
			bw_m->set(beg_idx, i, 0.0);
		}
		} else {
			model->transducer->labelsmap->get_column(&temp, seq[i].label);
			bw_m->column_multiply(i, &temp, 0);
		}
		
		scale[i] = scaling ? bw_m->sum_column(i) : 1;
		bw_m->column_multiply(i, 1/scale[i]);
	}
	if (bos)
		this->Zx = bw_m->get(beg_idx, 0);
	else
		{
			mathlib::DoubleVector temp(num_states);
			model->transducer->transitions->get_row(&temp, beg_idx);
			bw_m->column_multiply(1, &temp, 1);
			this->Zx = temp.sum();
		}
	
	//std::cout << Zx << std::endl;
}

void ForwardBackward::start_forward_scan()
{	
	*fw = 0.0;
	*next_fw = 0.0;
	if (bos)
		next_fw->set(beg_idx, 1/scale[0]);
	else next_fw->set(beg_idx, 1.0);
	index = 0;
	
}
void ForwardBackward::take_forward_step(data::Sequence & seq, models::Model * model, double * weights)
{
	index++;
	
	if (index < seq_len) {
	*fw = *next_fw;
	
	seq[index].compute_state_feature_scores(model->featureset, seq, index, this->Vi, weights);
	seq[index].compute_transition_feature_scores(model->featureset, this->Mi, weights);
	if (! eos) 
		Mi->set_column(end_idx, 1.0);
	if (! bos)
		Mi->set_row(beg_idx, 1.0);
	Mi->multiply_entrywise(model->transducer->transitions, 0);
	
	Mi->multiply(next_fw, fw, 1);
	next_fw->multiply(Vi);
	if (clamped)
		model->transducer->labelsmap->column_multiply(seq[index].label, next_fw, 1);
	next_fw->multiply(1/scale[index]);
	
	}
	
}
double ForwardBackward::current_forward(int state)
{	
	return next_fw->get(state);
}
double ForwardBackward::current_backward(int state)
{
	return bw_m->get(state, index);
}
double ForwardBackward::p(int state)
{
	double res;

	res = next_fw->get(state)*bw_m->get(state, index);
	
	res = scaling ? scale[index]*res : res / Zx;
	return res;
}
double ForwardBackward::p(int from_state, int to_state)
{
	double res;

	res = fw->get(from_state)*this->Mi->get(from_state, to_state)*this->Vi->get(to_state)*bw_m->get(to_state,index);
	res = scaling ? res : res / Zx;
	return res;
}

double ForwardBackward::p_i(int state, int pos)
{
	return bw_m->get(state, pos);
}

double ForwardBackward::get_partition_function()
{
	return this->Zx;
}

double ForwardBackward::compute_scaling_adjustment()
{
	int i;
	double res = 0.0;
	for (i=0; i<seq_len; i++)
		res += log(this->scale[i]);
	//std::cout << res << std::endl;
	return res;
}

void ForwardBackward::set_beg_idx(int beg_idx)
{
	this->beg_idx = beg_idx;
}

void ForwardBackward::set_end_idx(int end_idx)
{
	this->end_idx = end_idx;
}
}
