/*
 * Copyright (C) 2009 Castrense Savojardo
 * 	Bologna Biocomputing Group
 * 	University of Bologna, Italy
 *  savojard@biocomp.unibo.it
 * 
 * TrainerDG.cpp - This file is part of BioCRF
 * 
 * BioCRF is a free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 3 of the License,
 * or (at your option) any later version.
 *
 * BioCRF is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with BioCRF; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 * 
 */
#include <iostream>
#include <cmath>
#include <cstdlib>
#include "Model.h"
#include "Data.h"
#include "TrainerDG.h"
#include "DoubleMatrix.h"
#include "DoubleVector.h"
#include "Transducer.h"
#include "FeatureSet.h"
#include "Feature.h"
#include "ForwardBackward.h"

namespace training {

TrainerDG::TrainerDG(models::Model * model)
{
	this->model = model;
	fb_free = new algo::ForwardBackward(1, this->model->transducer->num_states);
	fb_clamped = new algo::ForwardBackward(1, this->model->transducer->num_states);
	fb_clamped->clamped = true;
	fb_clamped->scaling =  model->options->is_scaling;
	fb_free->scaling =  model->options->is_scaling;
}

TrainerDG::~TrainerDG()
{
	if (fb_free != NULL)
		delete fb_free;
	if (fb_clamped != NULL)
		delete fb_clamped;
}

void TrainerDG::train_LCRF()
{
	std::cout << "Starting training Order 1 Linear-Chain CRF model" << std::endl;
	
}

double TrainerDG::compute_cond_log_likelihood(double * gradient, double * lambda)
{
	int i, k;
	
	double loglik;
	double Zyx, Zx;
	
	
	int num_features = this->model->featureset->num_features; 
	
	data::Dataset::iterator datait;
	
	double ExpYX[num_features];
	double ExpX[num_features];
	
	// initialization
	loglik = 0.0;
	
	Zyx = 0.0;
	Zx = 0.0;
	for (i=0; i<num_features; i++) {
		loglik -= (lambda[i] * lambda[i]) / (2 * this->model->options->sigma_square);
		gradient[i] = -1*lambda[i] / this->model->options->sigma_square;
	}
	
	// compute log-likelihood and gradient vector over the whole dataset
	for (datait = this->model->data->ds->begin() ; 
		datait != this->model->data->ds->end(); datait++) {
	
		int seqlen = datait->size();
		
		for (i=0; i<num_features; i++){
			ExpYX[i] = 0.0;
			ExpX[i] = 0.0;
		}
		
		fb_free->run_backward(*datait, this->model, lambda);
		fb_clamped->run_backward(*datait, this->model, lambda);
		
		fb_free->start_forward_scan();
		fb_clamped->start_forward_scan();
		
		
		
		for (i=0; i<seqlen; i++) {

			std::vector<int>::iterator fit;
			for (fit = (*datait)[i].features.begin(); 
						fit != (*datait)[i].features.end(); ++fit) {
			/*
			for (boost::dynamic_bitset<>::size_type k = 0; k < (*datait)[i].feat_mask.size(); ++k){
				if ((*datait)[i].feat_mask[k]){
				*/
					if (this->model->featureset->features[*fit].type == STATE_FEATURE  && i>0 && i<seqlen-1){
						
						ExpX[*fit] += fb_free->p(this->model->featureset->features[*fit].to_state)*
							atof((*datait)[i+model->featureset->features[*fit].window_offset].cp[model->featureset->features[*fit].cp].c_str());
						ExpYX[*fit] += fb_clamped->p(this->model->featureset->features[*fit].to_state)*
							atof((*datait)[i+model->featureset->features[*fit].window_offset].cp[model->featureset->features[*fit].cp].c_str());
						
						/*
						ExpX[k] += fb_free->p(this->model->featureset->features[k].to_state)*
							(double)((*datait)[i+model->featureset->features[k].window_offset].cp[model->featureset->features[k].cp]/100.0);
						ExpYX[k] += fb_clamped->p(this->model->featureset->features[k].to_state)*
							(double)((*datait)[i+model->featureset->features[k].window_offset].cp[model->featureset->features[k].cp]/100.0);
							*/
					}
					
					if (this->model->featureset->features[*fit].type == TRANS_FEATURE && i > 0) {
						ExpX[*fit] += fb_free->p(model->featureset->features[*fit].from_state, model->featureset->features[*fit].to_state);
						ExpYX[*fit] += fb_clamped->p(model->featureset->features[*fit].from_state, model->featureset->features[*fit].to_state);
					}
			}
			
			/*
			
			std::map<int,double>::iterator fit;
			for (fit = (*datait)[i].features.begin(); 
						fit != (*datait)[i].features.end(); ++fit) {
				int type = this->model->featureset->features[fit->first].type;
				int from_state = this->model->featureset->features[fit->first].from_state;
				int to_state = this->model->featureset->features[fit->first].to_state;
				
				if (type == STATE_FEATURE) {
					ExpX[fit->first] += fb_free->p(to_state) * fit->second;
					
					ExpYX[fit->first] += fb_clamped->p(to_state) * fit->second;
				}
				
				if (type == TRANS_FEATURE && i > 0) {
					ExpX[fit->first] += fb_free->p(from_state, to_state) * fit->second;
					
					ExpYX[fit->first] += fb_clamped->p(from_state, to_state) * fit->second;
				}
			}*/
		
			
			fb_free->take_forward_step(*datait, this->model, lambda);
			fb_clamped->take_forward_step(*datait, this->model, lambda);
		
		}
		
		
		Zyx = fb_clamped->get_partition_function();
		Zx  = fb_free->get_partition_function();
		
		loglik += log(Zyx) - log(Zx);
		
		loglik -= fb_free->compute_scaling_adjustment();
		loglik += fb_clamped->compute_scaling_adjustment();
		
		
		for (i=0; i<num_features; i++) {
			gradient[i] += ExpYX[i];
			gradient[i] -= ExpX[i];
		} 
		
	}
	
	std::cout << "Log-likelihood       : " << loglik << std::endl;
	std::cout << "Gradient vector norm : " << norm(num_features, gradient) << std::endl;
	std::cout << "Lambda vector norm   : " << norm(num_features, lambda) << std::endl;
	
	return loglik;
}

void TrainerDG::train_GRCRF()
{
	int i;
	
	int iter;
	
	int iterations = this->model->options->iterations;
	int num_features = this->model->featureset->num_features;
	
	double loglik;
	double * gradient = new double[num_features];
	double * lambda = new double[num_features];
	
	iter = 1;
	
	for (i=0; i<num_features; i++){
		if (model->options->isretraining)
			lambda[i] = model->featureset->features[i].weight;
		else lambda[i] = this->model->options->init_params_value;
	}
	
	training::Optimizer * optmzr = this->model->get_optimizer();
	
	int iflag = 0;
	std::cout << "Starting training Grammatical-Renstrained CRF model" << std::endl;
	
	do {
		
		std::cout << "Iteration " << iter << std::endl;
		loglik = compute_cond_log_likelihood(gradient, lambda);

		/* negate function and gradient because optimizer minimizes 
	 	* the objective while we want maximizes it 
	 	*/
		loglik *= -1;
		for (i=0; i<num_features; i++)
			gradient[i] *= -1;
		
		optmzr->fval = loglik;
		optmzr->gradient = gradient;
		optmzr->parameters = lambda;
		
		iflag=optmzr->optimize();
		
		if (iflag < 0) {
			std::cout << "Optimization procedure encounters an error" << std::endl;
			break;
		}
		
		iter++;
		
	} while(iter <= iterations && iflag != 0);
	
	for (i=0; i<num_features; i++) {
		model->featureset->features[i].weight = lambda[i];
	}
	
	delete [] gradient;
	delete [] lambda;
	
	gradient = NULL;
	lambda = NULL;
	
	delete optmzr;
}

}
